@@ -72,12 +72,13 @@
 
 #ifdef __CYGWIN__
 #include <sys/cygwin.h>
+#include <windows.h>
 static int cygwin_version;
 static int get_cygwin_version()
 {
     int cygwin_version = 3004006;
     char *ver = (char *) cygwin_internal(CW_GETVERSIONINFO);
-	int major, minor;
+    int major, minor;
     if (2 == sscanf(ver, "%*s\n%%%%%% Cygwin dll major: %d\n"
                     "%%%%%% Cygwin dll minor: %d", &major, &minor))
         cygwin_version = major * 1000 + minor;
@@ -214,11 +215,14 @@ static void trigger(struct userdata *u, pa_sink_state_t sink_state, pa_source_st
     } else {
 
 #ifdef __CYGWIN__
-      if (cygwin_version > 3004006)
+        if (cygwin_version >= 3005000) {
 #endif
         if (enable_bits)
             if (ioctl(u->fd, SNDCTL_DSP_POST, NULL) < 0)
                 pa_log_warn("SNDCTL_DSP_POST: %s", pa_cstrerror(errno));
+#ifdef __CYGWIN__
+        }
+#endif
 
         if (!quick) {
             /*
@@ -913,7 +917,7 @@ static void thread_func(void *userdata) {
     int write_type = 0, read_type = 0;
     short revents = 0;
 #ifdef __CYGWIN__
-    int wait_usec;
+    ssize_t balance = 0; /* in usec */
 #endif
 
     pa_assert(u);
@@ -962,6 +966,7 @@ static void thread_func(void *userdata) {
                     } else {
                         l = info.bytes;
 #ifdef __CYGWIN__
+                        /* Reduce latency in cygwin before 3.5.0. */
                         if (cygwin_version < 3005000)
                             l -= info.fragsize * (info.fragstotal - 2);
 #endif
@@ -973,19 +978,13 @@ static void thread_func(void *userdata) {
                     }
                 }
 
+#ifndef __CYGWIN__ /* Cygwin does not need this. */
                 /* Round down to multiples of the fragment size,
                  * because OSS needs that (at least some versions
                  * do) */
                 l = (l/(ssize_t) u->out_fragment_size) * (ssize_t) u->out_fragment_size;
-
-#ifdef __CYGWIN__
-                wait_usec = 0;
-                if (l < 0)
-                    wait_usec = pa_bytes_to_usec(-l, &u->sink->sample_spec);
-                if (wait_usec > 10000)
-                    usleep(wait_usec);
-                if (cygwin_version >= 3005000) {
 #endif
+
                 /* Hmm, so poll() signalled us that we can read
                  * something, but GETOSPACE told us there was nothing?
                  * Hmm, make the best of it, try to read some data, to
@@ -994,9 +993,6 @@ static void thread_func(void *userdata) {
                     l = (ssize_t) u->out_fragment_size;
                     loop = false;
                 }
-#ifdef __CYGWIN__
-                }
-#endif
 
                 while (l > 0) {
                     void *p;
@@ -1113,6 +1109,11 @@ static void thread_func(void *userdata) {
                     p = pa_memblock_acquire(memchunk.memblock);
                     t = pa_read(u->fd, p, k, &read_type);
                     pa_memblock_release(memchunk.memblock);
+#ifdef __CYGWIN__
+                    /* Workaround for a bug of cygwin till 3.4.8 */
+                    if (t == UINT_MAX && cygwin_version <= 3004008)
+                        t = -1;
+#endif
 
                     pa_assert(t != 0); /* EOF cannot happen */
 
@@ -1204,6 +1205,57 @@ static void thread_func(void *userdata) {
                 goto fail;
             }
         }
+
+#ifdef __CYGWIN__
+        /* select()/poll() do not work in cygwin before 3.5.0 */
+        /* Emulate rtpoll stuff. */
+        if (cygwin_version < 3005000) {
+            const ssize_t sleep_time = 500; /* usec */
+            ssize_t allowable_wait_time = 5000; /* usec */
+            struct pollfd *pollfd;
+
+            if (u->rtpoll_item) {
+                pollfd = pa_rtpoll_item_get_pollfd(u->rtpoll_item, NULL);
+                revents = pollfd->revents;
+                revents &= ~POLLHUP;
+            } else
+                revents = 0;
+
+
+            while (allowable_wait_time > 0) {
+                struct audio_buf_info info;
+
+                if (!(u->sink && PA_SINK_IS_OPENED(u->sink->thread_info.state)) && !(u->source && PA_SOURCE_IS_OPENED(u->source->thread_info.state))) {
+                    revents &= ~(POLLOUT | POLLIN);
+                    break;
+                }
+                if (revents & POLLOUT) {
+                    if (ioctl(u->fd, SNDCTL_DSP_GETOSPACE, &info) < -1)
+                        goto fail;
+                    if (info.bytes < info.fragsize * (info.fragstotal - 2))
+                        revents &= ~POLLOUT;
+                }
+                if (revents & POLLIN) {
+                    if (ioctl(u->fd, SNDCTL_DSP_GETISPACE, &info) < -1)
+                        goto fail;
+                    if (info.bytes == 0)
+                        revents &= ~POLLIN;
+                }
+                if (revents)
+                    break;
+                balance += sleep_time;
+                allowable_wait_time -= sleep_time;
+                if (balance > sleep_time) {
+                    struct timespec tv0, tv1;
+                    clock_gettime(CLOCK_MONOTONIC, &tv0);
+                    usleep(sleep_time);
+                    clock_gettime(CLOCK_MONOTONIC, &tv1);
+                    balance -= (tv1.tv_sec * 1000000 + tv1.tv_nsec / 1000)
+                        - (tv0.tv_sec * 1000000 + tv0.tv_nsec / 1000);
+                }
+            }
+        }
+#endif
     }
 
 fail:
